
open Abstract
open Utils


let run_example i (p, sigma, descr) =
    print_string ("* Running test " ^ string_of_int i ^ ": " ^ descr ^ ".\n") ;
    print_string "** Program\n" ;
    print_string "#+BEGIN_SRC\n" ;
    print_string (string_of_prog p ^ "\n") ;
    print_string "#+END_SRC\n" ;
    print_string ("On context:\n" ^ string_of_sigma sigma ^ ".\n\n") ;
    List.iter (fun (analyser, str) ->
        print_string ("** Running analyser: =" ^ str ^ "=.") ;
        flush stdout ;
        analyser_match (analyser p sigma)
          (fun r ->
            print_string ("\t(Success)\nTerminated with result:\t" ^ string_of_ares r ^ "\n"))
          (fun lf lo ->
            let cps = compare_lex compare (compare_of_order (PosetDec.dec sign_ast_Poset)) in
            let cpsr = compare_lex cps (compare_of_order (PosetDec.dec sign_ares_Poset)) in
            print_string "\t(Failure)\nFailed to get a result.\n" ;
            List.iter (fun (p, sigma) ->
                print_string "*** Failed at\n" ;
                print_string "#+BEGIN_SRC\n" ;
                print_string (string_of_prog p ^ "\n") ;
                print_string ("On context:\n" ^ string_of_sigma sigma ^ ".\n") ;
                print_string "#+END_SRC\n")
              (normalise cps lf) ;
            List.iter (fun ((p, sigma), r) ->
                print_string "*** Oracle hint\n" ;
                print_string "#+BEGIN_SRC\n" ;
                print_string (string_of_prog p ^ "\n") ;
                print_string ("On context:\n" ^ string_of_sigma sigma ^ ".\n") ;
                print_string ("With result:\n" ^ string_of_ares r ^ ".\n") ;
                print_string "#+END_SRC\n")
              (normalise cpsr lo))
        ) [
            sign_trivial_analyser, "sign_trivial_analyser" ;
            sign_type_analyse, "sign_type_analyse" ;
            sign_analyse_fuel_fail 4, "sign_analyse_fuel_fail 4" ;
            sign_analyse_fuel_type 4, "sign_analyse_fuel type_analyse 4" ;
            sign_analyse_fuel_fail 100, "sign_analyse_fuel_fail 100" ;
            sign_analyse_fuel_type 100, "sign_analyse_fuel type_analyse 100" ;
            sign_analyse_deep_fuel_fail 1, "sign_analyse_deep_fuel_fail 1" ;
            sign_analyse_deep_fuel_type 1, "sign_analyse_deep_fuel type_analyse 1" ;
            sign_analyse_deep_fuel_fail 10, "sign_analyse_deep_fuel_fail 10" ;
            sign_analyse_deep_fuel_type 10, "sign_analyse_deep_fuel type_analyse 10" ;
            sign_analyse_oracle_back_and_forth 10 1 2,
            "sign_analyse_oracle_back_and_forth 10 1 2" ;
            sign_analyse_oracle_back_and_forth 50 1 2,
            "sign_analyse_oracle_back_and_forth 50 1 2" ;
            sign_analyse_oracle_back_and_forth 10 3 2,
            "sign_analyse_oracle_back_and_forth 10 3 2" ;
            sign_analyse_oracle_back_and_forth 50 3 2,
            "sign_analyse_oracle_back_and_forth 50 3 2" ;
            sign_analyse_oracle_back_and_forth 10 1 10,
            "sign_analyse_oracle_back_and_forth 10 1 10" ;
            sign_analyse_oracle_back_and_forth 50 1 10,
            "sign_analyse_oracle_back_and_forth 50 1 10" ;
            sign_analyse_oracle_back_and_forth 10 3 10,
            "sign_analyse_oracle_back_and_forth 10 3 10" ;
            sign_analyse_oracle_back_and_forth 50 3 10,
            "sign_analyse_oracle_back_and_forth 50 3 10" ;
            sign_analyse_oracle_back_and_forth_loop 10 1 2,
            "sign_analyse_oracle_back_and_forth_loop 10 1 2" ;
            sign_analyse_oracle_back_and_forth_loop 50 1 2,
            "sign_analyse_oracle_back_and_forth_loop 50 1 2" ;
            sign_analyse_oracle_back_and_forth_loop 10 3 2,
            "sign_analyse_oracle_back_and_forth_loop 10 3 2" ;
            sign_analyse_oracle_back_and_forth_loop 50 3 2,
            "sign_analyse_oracle_back_and_forth_loop 50 3 2" ;
            sign_analyse_oracle_back_and_forth_loop 10 1 10,
            "sign_analyse_oracle_back_and_forth_loop 10 1 10" ;
            sign_analyse_oracle_back_and_forth_loop 50 1 10,
            "sign_analyse_oracle_back_and_forth_loop 50 1 10" ;
            sign_analyse_oracle_back_and_forth_loop 10 3 10,
            "sign_analyse_oracle_back_and_forth_loop 10 3 10" ;
            sign_analyse_oracle_back_and_forth_loop 50 3 10,
            "sign_analyse_oracle_back_and_forth_loop 50 3 10" ;
            sign_analyse_oracle_back_and_forth_deep 1 1 1 2,
            "sign_analyse_oracle_back_and_forth_deep 1 1 1 2" ;
            sign_analyse_oracle_back_and_forth_deep 1 5 1 2,
            "sign_analyse_oracle_back_and_forth_deep 1 5 1 2" ;
            sign_analyse_oracle_back_and_forth_deep 1 1 3 2,
            "sign_analyse_oracle_back_and_forth_deep 1 1 3 2" ;
            sign_analyse_oracle_back_and_forth_deep 1 5 3 2,
            "sign_analyse_oracle_back_and_forth_deep 1 5 3 2" ;
            sign_analyse_oracle_back_and_forth_deep 1 1 1 10,
            "sign_analyse_oracle_back_and_forth_deep 1 1 1 10" ;
            sign_analyse_oracle_back_and_forth_deep 1 5 1 10,
            "sign_analyse_oracle_back_and_forth_deep 1 5 1 10" ;
            sign_analyse_oracle_back_and_forth_deep 1 1 3 10,
            "sign_analyse_oracle_back_and_forth_deep 1 1 3 10" ;
            sign_analyse_oracle_back_and_forth_deep 1 5 3 10,
            "sign_analyse_oracle_back_and_forth_deep 1 5 3 10" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 1 1 2,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 1 1 2" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 5 1 2,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 5 1 2" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 1 3 2,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 1 3 2" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 5 3 2,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 5 3 2" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 1 1 10,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 1 1 10" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 5 1 10,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 5 1 10" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 1 3 10,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 1 3 10" ;
            sign_analyse_oracle_back_and_forth_deep_loop 1 5 3 10,
            "sign_analyse_oracle_back_and_forth_deep_loop 1 5 3 10"
        ] ;
    print_newline ()

let _ =
    print_string "#+INFOJS_OPT: view:overview toc:nil\n\n" ;
    print_string "Running the analysers on different tests.\n\n" ;
    let x_defined = write_env initial ['x'] Sign.Coq_tops in
    List.iteri run_example [
        prog_test0, Sign_ast_prog initial, "6 × 7 using a loop" ;
        prog_test1, Sign_ast_prog initial, "an assignment" ;
        prog_test2, Sign_ast_prog x_defined, "limitation of the absence of meet after branching" ;
        prog_test3, Sign_ast_prog x_defined, "a loop for which the signs does not fit" ;
        prog_test4, Sign_ast_prog x_defined, "an unexecuted loop" ;
        prog_test5, Sign_ast_prog x_defined, "a neverending loop" ;
        prog_test6, Sign_ast_prog initial, "a relatively long computation without loop" ;
        prog_test7, Sign_ast_prog initial, "a loop terminating after two unfoldings" ;
        prog_test8, Sign_ast_prog initial, "a loop terminating after four unfoldings" ;
        prog_test9, Sign_ast_prog initial, "testing [if] and [throw]" ;
        prog_test10, Sign_ast_prog initial, "a loop terminating after four unfoldings with tests" ;
        prog_test11, Sign_ast_prog x_defined, "nested loops" ;
        prog_test12, Sign_ast_prog initial, "a loop whose termination is difficult to tell"
    ]


