
open Abstract
open List


let rec uniq cmp = function
    | [] -> []
    | a :: b :: l when cmp a b = 0 -> uniq cmp (b :: l)
    | a :: l -> a :: uniq cmp l

let normalise compare l =
    uniq compare (List.sort compare l)

let compare_lex o1 o2 (a, b) (x, y) =
    let r1 = o1 a x in
    if r1 = 0 then o2 b y else r1

let compare_of_order o a b =
    if o a b then
        if o b a then 0
        else -1
    else 1

let par s = "(" ^ s ^ ")"
let ret n = "\n" ^ String.make n ' '

let string_of_var x =
    String.concat "" (map (String.make 1) x)

let rec string_of_expr n = function
    | Expr_cst c -> par (string_of_int c)
    | Expr_var x -> par (string_of_var x)
    | Expr_add (e1, e2) -> par (string_of_expr n e1 ^ " + " ^ string_of_expr n e2)

let string_of_ext_expr n = function
    | Expr_add_1 e -> "add1 " ^ string_of_expr (n + 5) e
    | Expr_add_2 -> "add2"

let rec string_of_stat n = function
    | Stat_skip -> "skip"
    | Stat_asgn (x, e) -> string_of_var x ^ " := " ^ string_of_expr (n + length x + 4) e
    | Stat_seq (s1, s2) -> string_of_stat n s1 ^ " ;" ^ ret n ^ string_of_stat n s2
    | Stat_if (e, s1, s2) ->
        "if " ^ string_of_expr (n + 3) e ^ ret n ^ " then " ^ string_of_stat (n + 6) s1 ^ ret n ^ " else " ^ string_of_stat (n + 6) s2 ^ ret n ^ "fi"
    | Stat_while (e, s) -> "while " ^ string_of_expr (n + 6) e ^ ret n ^ "do " ^ string_of_stat (n + 3) s ^ ret n ^ "od"
    | Stat_throw -> "throw"

let string_of_ext_stat n = function
    | Stat_asgn_1 x -> string_of_var x ^ " (:=)1"
    | Stat_seq_1 s -> "(;)1" ^ ret n ^ string_of_stat n s
    | Stat_if_1 (s1, s2) -> "if1" ^ ret n ^ " then " ^ string_of_stat (n + 6) s1 ^ ret n ^ " else " ^ string_of_stat (n + 6) s2 ^ ret n ^ "fi"
    | Stat_while_1 (e, s) -> "while1 " ^ string_of_expr (n + 7) e ^ ret n ^ "do " ^ string_of_stat (n + 3) s ^ ret n ^ "od"
    | Stat_while_2 (e, s) -> "while2 " ^ string_of_expr (n + 7) e ^ ret n ^ "do " ^ string_of_stat (n + 3) s ^ ret n ^ "od"

let string_of_prog = function
    | Prog_expr e -> "[ " ^ string_of_expr 2 e ^ " ]"
    | Prog_ext_expr e -> "[ " ^ string_of_ext_expr 2 e ^ " ]"
    | Prog_stat s -> "[ " ^ string_of_stat 2 s ^ " ]"
    | Prog_ext_stat s -> "[ " ^ string_of_ext_stat 2 s ^ " ]"

let string_of_err e =
    if e Err then "<may be an error>"
    else "<no error>"

let string_of_val = function
    | Sign.Coq_bot -> "Bot"
    | Sign.Coq_tops -> "Top (sign)"
    | Sign.Coq_pos -> "+"
    | Sign.Coq_neg -> "-"
    | Sign.Coq_zero -> "0"
    | Sign.Coq_pos0 -> "+0"
    | Sign.Coq_neg0 -> "-0"
    | Sign.Coq_not0 -> "+-"

let string_of_undef = function
    | None -> "Bot"
    | Some () -> "Undef"


let string_of_aValVar = function
    | None -> "Top (valvar)"
    | Some (Inl v) -> string_of_val v
    | Some (Inr u) -> string_of_undef u

let string_of_env (d, m) =
    "{ " ^ (
      fold var_OrderedType (mapAVL_FMap var_OrderedType) (fun x v str ->
        if eq_aValVar v d then str
        else string_of_var x ^ " -> " ^ string_of_aValVar v ^ ", " ^ str) m ""
    ) ^ "_ -> " ^ string_of_aValVar d
    ^ " }"

let string_of_ares = function
    | Sign_ares_top -> "Top"
    | Sign_ares_bot -> "Bot"
    | Sign_ares_expr (v, e) -> "expr (" ^ string_of_val v ^ " , " ^ string_of_err e ^ ")"
    | Sign_ares_stat (env, e) -> "stat (" ^ string_of_env env ^ " , " ^ string_of_err e ^ ")"

let string_of_sigma = function
    | Sign_ast_prog e -> string_of_env e
    | Sign_ast_add_1 (e, r) -> "add1 (" ^ string_of_env e ^ " , " ^ string_of_ares r ^ ")"
    | Sign_ast_add_2 (v, r) -> "add2 (" ^ string_of_val v ^ " , " ^ string_of_ares r ^ ")"
    | Sign_ast_asgn_1 (e, r) -> "asgn1 (" ^ string_of_env e ^ " , " ^ string_of_ares r ^ ")"
    | Sign_ast_seq_1 r -> "seq1 (" ^ string_of_ares r ^ ")"
    | Sign_ast_if_1 (e, r) -> "if1 (" ^ string_of_env e ^ " , " ^ string_of_ares r ^ ")"
    | Sign_ast_while_1 r -> "while1 (" ^ string_of_ares r ^ ")"
    | Sign_ast_while_2 (e, r) -> "while2 (" ^ string_of_env e ^ " , " ^ string_of_ares r ^ ")"
    | Sign_ast_top -> "Top"
    | Sign_ast_bot -> "Bot"

